# Copyright (C) 2010 by Dr. Dieter Maurer <dieter@handshake.de>; see 'LICENSE.txt' for details
"""Support for the http artifact binding.

The artifact binding allows to transport a small reference to an
SAML2 message, called ``artifact`` via
via parameters in an HTTP ``GET`` or ``POST``
request.

The artifact can later be derefenced via the artifact resolution
protocol.

Artifact handling is supported by an ``ArtifactManager``,
defined in module ``util``.
"""
from dm.saml2.uuid import UUID

def encode(message, manager, relay_state=None):
  """encode *message* with help of *manager*; return a params dict.

  *message* is a ``PyXB`` binding of an SAML message.

  ``urllib.urlencode`` can be used to prepare the returned params
  for use as url parameters and the ``to_controls`` and ``to_form``
  functions in ``httppost`` to use them in a form.
  """
  params = {}
  if relay_state: params['RelayState'] = relay_state
  epi = manager.index
  EndpointIndex = chr(epi//256) + chr(epi % 256)
  TypeCode='\0x0\0x4'
  uuid = UUID(manager.store(message.toxml())).bytes + '\0\0\0\0'
  RemainingArtifact = manager.id + uuid
  artifact = (TypeCode + EndpointIndex + RemainingArtifact).encode('base64')
  # strip whitespace
  artifact = ''.join(artifact.split())
  params['SAMLart'] = artifact
  return params


def parse(params):
  """parse *params* into *SourceID*, *EndpointIndex*, *artifact* and *relay_state*.

  Used by artifact receiver to obtain the information necessary
  for artifact resolution.
  """
  relay_state = params.get('RelayState')
  artifact = params['SAMLart']
  return parse_artifact(artifact)[:2] + (artifact, relay_state)


def parse_artifact(artifact):
  """parse artifact into *SourceID*, *EndpointIndex* and *MessageHandle*."""
  artifact = artifact.decode('base64')
  if artifact[:2] != '\0x0\0x4':
    raise NotImplementedError('unsupported artifact type: %s' % artifact[:2])
  EndpointIndex = ord(artifact[2])*256 + ord(artifact[3])
  SourceID = artifact[4:24]
  MessageHandle = artifact[24:]
  return EndpointIndex, SourceID, MessageHandle


def resolve(artifact, manager):
  """resolve *artifact* into associated SAML2 message using *manager*.
  Note: the SAML2 message is returned as a document (string), not as
  its ``PyXB`` binding.

  Raise ``KeyError`` in case of resolution problems.
  The ``KeyError`` may be the subtype ``dm.saml2.binding.util.TimeoutError``
  to indicate that the artifact could be resolved but had timed out.
  """
  handle = parse_artifact(artifact)[2]
  uuid = UUID(bytes=handle[:16])
  return manager.retrieve(uuid)



"""id generation for saml2."""


try: from uuid import uuid4
except ImportError:
  # code copied from Python's "uuid" module (for Python 2.4 compatibility)
  class UUID(object):
    """Instances of the UUID class represent UUIDs as specified in RFC 4122.
    UUID objects are immutable, hashable, and usable as dictionary keys.
    Converting a UUID to a string with str() yields something in the form
    '12345678-1234-1234-1234-123456789abc'.  The UUID constructor accepts
    five possible forms: a similar string of hexadecimal digits, or a tuple
    of six integer fields (with 32-bit, 16-bit, 16-bit, 8-bit, 8-bit, and
    48-bit values respectively) as an argument named 'fields', or a string
    of 16 bytes (with all the integer fields in big-endian order) as an
    argument named 'bytes', or a string of 16 bytes (with the first three
    fields in little-endian order) as an argument named 'bytes_le', or a
    single 128-bit integer as an argument named 'int'.

    UUIDs have these read-only attributes:

        bytes       the UUID as a 16-byte string (containing the six
                    integer fields in big-endian byte order)

        bytes_le    the UUID as a 16-byte string (with time_low, time_mid,
                    and time_hi_version in little-endian byte order)

        fields      a tuple of the six integer fields of the UUID,
                    which are also available as six individual attributes
                    and two derived attributes:

            time_low                the first 32 bits of the UUID
            time_mid                the next 16 bits of the UUID
            time_hi_version         the next 16 bits of the UUID
            clock_seq_hi_variant    the next 8 bits of the UUID
            clock_seq_low           the next 8 bits of the UUID
            node                    the last 48 bits of the UUID

            time                    the 60-bit timestamp
            clock_seq               the 14-bit sequence number

        hex         the UUID as a 32-character hexadecimal string

        int         the UUID as a 128-bit integer

        urn         the UUID as a URN as specified in RFC 4122

        variant     the UUID variant (one of the constants RESERVED_NCS,
                    RFC_4122, RESERVED_MICROSOFT, or RESERVED_FUTURE)

        version     the UUID version number (1 through 5, meaningful only
                    when the variant is RFC_4122)
    """

    def __init__(self, hex=None, bytes=None, bytes_le=None, fields=None,
                       int=None, version=None):
        r"""Create a UUID from either a string of 32 hexadecimal digits,
        a string of 16 bytes as the 'bytes' argument, a string of 16 bytes
        in little-endian order as the 'bytes_le' argument, a tuple of six
        integers (32-bit time_low, 16-bit time_mid, 16-bit time_hi_version,
        8-bit clock_seq_hi_variant, 8-bit clock_seq_low, 48-bit node) as
        the 'fields' argument, or a single 128-bit integer as the 'int'
        argument.  When a string of hex digits is given, curly braces,
        hyphens, and a URN prefix are all optional.  For example, these
        expressions all yield the same UUID:

        UUID('{12345678-1234-5678-1234-567812345678}')
        UUID('12345678123456781234567812345678')
        UUID('urn:uuid:12345678-1234-5678-1234-567812345678')
        UUID(bytes='\x12\x34\x56\x78'*4)
        UUID(bytes_le='\x78\x56\x34\x12\x34\x12\x78\x56' +
                      '\x12\x34\x56\x78\x12\x34\x56\x78')
        UUID(fields=(0x12345678, 0x1234, 0x5678, 0x12, 0x34, 0x567812345678))
        UUID(int=0x12345678123456781234567812345678)

        Exactly one of 'hex', 'bytes', 'bytes_le', 'fields', or 'int' must
        be given.  The 'version' argument is optional; if given, the resulting
        UUID will have its variant and version set according to RFC 4122,
        overriding the given 'hex', 'bytes', 'bytes_le', 'fields', or 'int'.
        """

        if [hex, bytes, bytes_le, fields, int].count(None) != 4:
            raise TypeError('need one of hex, bytes, bytes_le, fields, or int')
        if hex is not None:
            hex = hex.replace('urn:', '').replace('uuid:', '')
            hex = hex.strip('{}').replace('-', '')
            if len(hex) != 32:
                raise ValueError('badly formed hexadecimal UUID string')
            int = long(hex, 16)
        if bytes_le is not None:
            if len(bytes_le) != 16:
                raise ValueError('bytes_le is not a 16-char string')
            bytes = (bytes_le[3] + bytes_le[2] + bytes_le[1] + bytes_le[0] +
                     bytes_le[5] + bytes_le[4] + bytes_le[7] + bytes_le[6] +
                     bytes_le[8:])
        if bytes is not None:
            if len(bytes) != 16:
                raise ValueError('bytes is not a 16-char string')
            int = long(('%02x'*16) % tuple(map(ord, bytes)), 16)
        if fields is not None:
            if len(fields) != 6:
                raise ValueError('fields is not a 6-tuple')
            (time_low, time_mid, time_hi_version,
             clock_seq_hi_variant, clock_seq_low, node) = fields
            if not 0 <= time_low < 1<<32L:
                raise ValueError('field 1 out of range (need a 32-bit value)')
            if not 0 <= time_mid < 1<<16L:
                raise ValueError('field 2 out of range (need a 16-bit value)')
            if not 0 <= time_hi_version < 1<<16L:
                raise ValueError('field 3 out of range (need a 16-bit value)')
            if not 0 <= clock_seq_hi_variant < 1<<8L:
                raise ValueError('field 4 out of range (need an 8-bit value)')
            if not 0 <= clock_seq_low < 1<<8L:
                raise ValueError('field 5 out of range (need an 8-bit value)')
            if not 0 <= node < 1<<48L:
                raise ValueError('field 6 out of range (need a 48-bit value)')
            clock_seq = (clock_seq_hi_variant << 8L) | clock_seq_low
            int = ((time_low << 96L) | (time_mid << 80L) |
                   (time_hi_version << 64L) | (clock_seq << 48L) | node)
        if int is not None:
            if not 0 <= int < 1<<128L:
                raise ValueError('int is out of range (need a 128-bit value)')
        if version is not None:
            if not 1 <= version <= 5:
                raise ValueError('illegal version number')
            # Set the variant to RFC 4122.
            int &= ~(0xc000 << 48L)
            int |= 0x8000 << 48L
            # Set the version number.
            int &= ~(0xf000 << 64L)
            int |= version << 76L
        self.__dict__['int'] = int

    def __cmp__(self, other):
        if isinstance(other, UUID):
            return cmp(self.int, other.int)
        return NotImplemented

    def __hash__(self):
        return hash(self.int)

    def __int__(self):
        return self.int

    def __repr__(self):
        return 'UUID(%r)' % str(self)

    def __setattr__(self, name, value):
        raise TypeError('UUID objects are immutable')

    def __str__(self):
        hex = '%032x' % self.int
        return '%s-%s-%s-%s-%s' % (
            hex[:8], hex[8:12], hex[12:16], hex[16:20], hex[20:])

    def get_bytes(self):
        bytes = ''
        for shift in range(0, 128, 8):
            bytes = chr((self.int >> shift) & 0xff) + bytes
        return bytes

    bytes = property(get_bytes)

    def get_bytes_le(self):
        bytes = self.bytes
        return (bytes[3] + bytes[2] + bytes[1] + bytes[0] +
                bytes[5] + bytes[4] + bytes[7] + bytes[6] + bytes[8:])

    bytes_le = property(get_bytes_le)

    def get_fields(self):
        return (self.time_low, self.time_mid, self.time_hi_version,
                self.clock_seq_hi_variant, self.clock_seq_low, self.node)

    fields = property(get_fields)

    def get_time_low(self):
        return self.int >> 96L

    time_low = property(get_time_low)

    def get_time_mid(self):
        return (self.int >> 80L) & 0xffff

    time_mid = property(get_time_mid)

    def get_time_hi_version(self):
        return (self.int >> 64L) & 0xffff

    time_hi_version = property(get_time_hi_version)

    def get_clock_seq_hi_variant(self):
        return (self.int >> 56L) & 0xff

    clock_seq_hi_variant = property(get_clock_seq_hi_variant)

    def get_clock_seq_low(self):
        return (self.int >> 48L) & 0xff

    clock_seq_low = property(get_clock_seq_low)

    def get_time(self):
        return (((self.time_hi_version & 0x0fffL) << 48L) |
                (self.time_mid << 32L) | self.time_low)

    time = property(get_time)

    def get_clock_seq(self):
        return (((self.clock_seq_hi_variant & 0x3fL) << 8L) |
                self.clock_seq_low)

    clock_seq = property(get_clock_seq)

    def get_node(self):
        return self.int & 0xffffffffffff

    node = property(get_node)

    def get_hex(self):
        return '%032x' % self.int

    hex = property(get_hex)

    def get_urn(self):
        return 'urn:uuid:' + str(self)

    urn = property(get_urn)

    def get_variant(self):
        if not self.int & (0x8000 << 48L):
            return RESERVED_NCS
        elif not self.int & (0x4000 << 48L):
            return RFC_4122
        elif not self.int & (0x2000 << 48L):
            return RESERVED_MICROSOFT
        else:
            return RESERVED_FUTURE

    variant = property(get_variant)

    def get_version(self):
        # The version bits are only meaningful for RFC 4122 UUIDs.
        if self.variant == RFC_4122:
            return int((self.int >> 76L) & 0xf)

    version = property(get_version)

  def uuid4():
    """Generate a random UUID."""
    try:
        import os
        return UUID(bytes=os.urandom(16), version=4)
    except:
        import random
        bytes = [chr(random.randrange(256)) for i in range(16)]
        return UUID(bytes=bytes, version=4)

def uuid():
  return str(uuid4())




# Copyright (C) 2011-2012 by Dr. Dieter Maurer <dieter@handshake.de>
"""SAML2 SOAP binding."""

from logging import getLogger


class BadMessage(Exception):
  """Indicate a bad soap message."""


logger = getLogger(__name__)


SOAP_ENV_TEMPLATE="""<SOAP-ENV:Envelope xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/"><SOAP-ENV:Body>%s</SOAP-ENV:Body></SOAP-ENV:Envelope>""" 

# interestingly, the `Fault` children lack an XML namespace.
#  code taken from http://www.w3.org/TR/2000/NOTE-SOAP-20000508/#_Ref477795996
SOAP_FAULT_TEMPLATE="""<SOAP-ENV:Fault><faultcode>%s</faultcode><faultstring>%a</faultstring><detail><exception xmlns="urn:exception_info">%s</exception></detail></SOAP_ENV:Fault>""" 


def encapsulate(saml):
  """encapsulate *saml* with an SOAP envelope."""
  if hasattr(saml, "toxml"): saml = saml.toxml(root_only=True)
  if isinstance(saml, unicode): saml = saml.encode("utf-8")
  return SOAP_ENV_TEMPLATE % saml



def extract(soap):
  """extract the saml message from the SOAP envelope *soap*."""
  # we use libxml2 for parsing as it is needed for signing as well
  from libxml2 import parseMemory
  ps = parseMemory(soap, len(soap))
  try:
    r = ps.getRootElement()
    child = r.get_children() # this is indeed the first child
    while child is not None and child.get_name() != "Body": child = child.next
    # may raise `AttributeError: get_children` when *soap* is not valid
    message = child.get_children()
    # This must be an SAML `Request` or `Response` or an SOAP `Fault`.
    # We do not distinguish `Fault` and map it to `BadMessage`, as well
    if message.get_name() not in ("Request", "Response"):
      raise BadMessage(soap)
    message.reconciliateNs(ps)
    return message.serialize()
  finally: ps.freeDoc()
  


def http_post(transport, url, saml):
  """post *saml* message to *url* and return the *saml* response.

  *transport* is expected to have an `open` method with parameters *url*,
  *data* and *headers*. Note, that for real SAML2 compatibility, it must
  provide an HTTPS client certificate and verify the HTTPS server certificate.
  The result is expected to be similar to that of `urllib2.urlopen`;
  http errors are expected to raise an exception.
  """
  r = transport.open(
    url, encapsulate(saml),
    {"Cache-Control" : "no-cache, no-store",
     "Pragma" : "no-cache",
     "SOAPAction" : "http://www.oasis-open.org/committees/security",
     "Content-Type": "text/xml; charset=utf-8",
     }
    )
  return extract(r.read())
  
  
def http_request(soap, callback):
  """process *soap* request via *callback*.

  return a triple *body*, *headers*, *status* describing the result.

  Note, that if *status* is not 200, it might be necessary to
  abort a (potential) transaction.
  """
  try:
    result = callback(extract(soap), binding="SOAP")
  except Exception, e:
    logger.exception("processing %s failed", soap)
    # we do not distinguish between server and client errors
    # we always include `detail` (violating the SOAP 1.1 spec which
    #   request `detail` to be present only for `Body` processing problems
    result = SOAP_FAULT_TEMPLATE % (
      "SOAP-ENV:Server",
      "Server Error",
      "%s: %s" (e.__class__, e)
      )
    status = 500
  else: status = 200
  return (encapsulate(result),
          {"Cache-Control" : "no-cache, no-store, must-revalidate, private",
           "Pragma" : "no-cache"
           },
          status)
    



from pyxb.bundles.saml20.protocol import *

from dm.saml2.signature import add_signature_verification
from dm.saml2.util import enhance, Picklable

from assertion import _StandardSignable



class _RequestAbstractTypeExtension(_StandardSignable, Picklable):
  """A mix in class to provide extensions for ``RequestAbstractType``."""

class _ResponseTypeExtension(_StandardSignable):
  """A mix in class to provide extensions for ``StatusResponseType``."""

  PREFIX = "urn:oasis:names:tc:SAML:2.0:status:"

  def set_status(self, code, message=None, detail=None):
    """set the ``Status`` element.

    *code* is either a string or a sequence of strings or a ``StatusCodeType``
    instance. It specifies
    the content of the nested ``StatusCode`` elements. If a
    string does not contain ``:``, then *PREFIX* is prepended.

    *message* is a string and specifies the ``StatusMessage``.

    *detail* must be a ``StatusDetailType`` instance.
    """
    if not isinstance(code, StatusCodeType):
      if isinstance(code, (str, unicode)): code = code,
      code, css = None, code
      for cs in css:
        if ':' not in cs: cs = self.PREFIX + cs
        c = StatusCode(Value=cs)
        if code is None: parent = code = c
        else: parent.StatusCode = c; parent = c
    # ``code`` is not a ``StatusCodeType`` instance
    self.Status = Status(code, message, detail)

  def set_success(self):
    """set success status."""
    self.set_status('Success')

  def set_exception(self, exc):
    """set a status for *exc*.

    *exc* is an unexpected exception and therefore classified as
    responder failure. Exceptions due to client failures should
    not be handled by this method.
    """
    msg = '%s: %s' % (exc.__class__.__name__, str(exc))
    self.set_status('Responder', msg)

##  # first level status codes
##  Success
##  Requester
##  Responder
##  VersionMismatch

##  # second level status codes
##  AuthnFailed
##  InvalidAttrNameOrValue
##  UnkownAttrProfile
##  InvalidNameIDPolicy
##  NoAuthnContext
##  NoAvailableIDP
##  NoPassive
##  PartialLogout
##  # there are more, but we do not expect we will use them

enhance('StatusResponseType', _ResponseTypeExtension)
enhance('ResponseType')

# assertion request and assertion query
enhance('AssertionIDRequestType', _RequestAbstractTypeExtension)
enhance('AuthnQueryType', _RequestAbstractTypeExtension)
enhance('AttributeQueryType', _RequestAbstractTypeExtension)
enhance('AuthzDecisionQueryType', _RequestAbstractTypeExtension)

# authentication request
enhance('AuthnRequestType', _RequestAbstractTypeExtension)

# artifact resolution
enhance('ArtifactResolveType', _RequestAbstractTypeExtension)
enhance('ArtifactResponseType', _ResponseTypeExtension)

# name id management
enhance('ManageNameIDRequestType', _RequestAbstractTypeExtension)

# single logout
enhance('LogoutRequestType', _RequestAbstractTypeExtension)

# name id mapping
enhance('NameIDMappingRequestType', _RequestAbstractTypeExtension)
enhance('NameIDMappingResponseType', _ResponseTypeExtension)



add_signature_verification()




# Copyright (C) 2010 by Dr. Dieter Maurer <dieter@handshake.de>; see 'LICENSE.txt' for details
"""Utilities"""
from time import time

from dm.saml2.uuid import uuid

class Store(object):
  """class to implement storage of objects under ids.

  Instances of this class are used to manage relay states and
  artifacts. They use a mapping store for the actual storage.

  Artifact management stores pairs consisting of a timestamp (a float)
  and an XML document. Relay state management stores whatever the
  application stores. The used mapping store must be able to store
  accordingly.

  In both cases, occational errors may lead to old (no longer used)
  data records. Therefore, it may become necessary to clear the
  store from time to time.
  """
  def __init__(self, mapping_store):
    self._store = mapping_store

  def store(self, obj):
    """store *obj* and return an id to retrieve it later."""
    id = self._makeid()
    self._store[id] = obj
    return id

  def retrieve(self, id):
    """retrieve the object stored under *id* and delete."""
    obj = self._store[id]
    del self._store[id]
    return obj

  @staticmethod
  def _makeid():
    """auxiliary method to generate ids."""
    return uuid()

  def clear(self): self._store.clear()


class UnmanagedError(Exception):
  """exception indicating that the object is not managed."""


class RelayStateManager(object):
  """manages relay states.

  delegates storage to a ``Store`` instance.
  """

  PREFIX = 'rs:' # indicates a managed relay state

  def __init__(self, store):
    self._store = store

  def store(self, state):
    return self.PREFIX + self._store.store(state)

  def retrieve(self, id):
    p = self.PREFIX
    if not id.startswith(p): raise UnmanagedError(id)
    return self._store.retrieve(id[len(p):])

  def clear(self): self._store.clear()


class TimeoutError(KeyError):
  """exception indicating that the value timed out."""


class ArtifactManager(object):
  """manages artifacts.

  stores XML documents for a limited time.
  """
  VALID_SECONDS = 300

  id = None # associated ``SourceID``
  index = 0 # assiciated ``EndpointIndex``

  def __init__(self, store, id, index=0, valid_seconds=None):
    """*id* specifies the ``SourceID``, it should be the SHA-1 hash of the entity name,
    *index* specifies the ``EndpointIndex`` (an integer).
    *valid_seconds* specifies how long artifacts can be resolved.

    *store* must be able to store pairs float, string.
    """
    self._store = store
    assert len(id) == 20 # may compute the SHA-1 hash ourselves
    self.id = id; self.index = index
    self._valid = valid_seconds > 0 or self.VALID_SECONDS

  def store(self, doc):
    return self._store.store((time(), obj))

  def retrieve(self, id):
    ts, obj = self._store.retrieve(id)
    if time() - ts > self._valid: raise TimeoutError(id, ts)
    return obj



